import unittest

from sqlalchemy import Unicode
from nanohttp import settings, json, etag, context

from restfulpy.orm import DeclarativeBase, DBSession, Field, ModifiedMixin, commit
from restfulpy.testing import WebAppTestCase, FormParameter
from restfulpy.tests.helpers import MockupApplication
from restfulpy.controllers import ModelRestController


class EtagCheckingModel(ModifiedMixin, DeclarativeBase):
    __tablename__ = 'etag_checking_model'

    title = Field(Unicode(50), primary_key=True)

    @property
    def __etag__(self):
        return self.last_modification_time.isoformat()


class Root(ModelRestController):
    __model__ = EtagCheckingModel

    @json
    @etag
    @commit
    def post(self):
        m = EtagCheckingModel()
        m.update_from_request()
        DBSession.add(m)
        return m

    @json
    @etag
    @EtagCheckingModel.expose
    def get(self, title: str=None):
        query = EtagCheckingModel.query
        if title:
            return query.filter(EtagCheckingModel.title == title).one_or_none()
        return query

    @json
    @etag
    @EtagCheckingModel.expose
    @commit
    def put(self, title: str=None):
        m = DBSession.query(EtagCheckingModel).filter(EtagCheckingModel.title == title).one_or_none()
        m.update_from_request()
        context.etag_match(m.__etag__)
        return m


class EtagCheckingModelTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root())
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_etag_match(self):
        resp, headers = self.request('ALL', 'POST', '/', params=[
            FormParameter('title', 'etag_test')
        ])
        self.assertIn('ETag', headers)
        initial_etag = headers['ETag']

        # Getting the resource with known etag, expected 304
        self.request(
            'ALL', 'GET', '/etag_test',
            headers={
                'If-None-Match': initial_etag
            },
            expected_status=304
        )

        # Putting without the etag header, expected error: Precondition Failed
        self.request(
            'ALL', 'PUT', '/etag_test',
            params=[
                FormParameter('title', 'etag_test_edit1')
            ],
            expected_status=412
        )

        # Putting with the etag header, expected: success
        resp, headers = self.request(
            'ALL', 'PUT', '/etag_test',
            params=[
                FormParameter('title', 'etag_test_edit1')
            ],
            headers={
                'If-Match': initial_etag
            }
        )
        self.assertIn('ETag', headers)
        etag_after_put = headers['ETag']
        self.assertNotEqual(initial_etag, etag_after_put)

        # Getting the resource with known etag, expected 304
        self.request(
            'ALL', 'GET', '/etag_test_edit1',
            headers={
                'If-None-Match': initial_etag
            },
            expected_status=200
        )
        self.assertIn('ETag', headers)
        new_etag = headers['ETag']
        self.assertNotEqual(new_etag, initial_etag)
        self.assertEqual(new_etag, etag_after_put)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




from nanohttp import HttpStatus, HttpInternalServerError


class RestfulException(Exception):
    def to_json(self):
        return {
            'type': str(type(self)),
            'content': str(self),
        }


class UnsupportedError(Exception):
    """Used to indicate we're not supporting a feature and or method.

    .. versionadded:: 0.11.0

    """
    pass


class SqlError(HttpStatus):

    def __init__(self, sqlalchemy_error):
        pgcode, self.status_code, self.status_text, message = self.map_exception(sqlalchemy_error)
        super().__init__(message, reason=pgcode)

    @classmethod
    def map_exception(cls, ex):
        if not hasattr(ex, 'orig') or ex.orig is None:
            return 500, HttpInternalServerError.status_text, ''

        error_code = ex.orig.pgcode
        status_text = cls.postgresql_errors.get(error_code)
        return error_code, cls.statuses.get(error_code, 500), status_text, ex.orig.pgerror

    statuses = {
        '23502': 400,
        '23505': 409
    }
    postgresql_errors = {
        '00000': 'successful_completion',
        '01000': 'warning',
        '0100C': 'dynamic_result_sets_returned',
        '01008': 'implicit_zero_bit_padding',
        '01003': 'null_value_eliminated_in_set_function',
        '01007': 'privilege_not_granted',
        '01006': 'privilege_not_revoked',
        '01004': 'string_data_right_truncation',
        '01P01': 'deprecated_feature',
        '02000': 'no_data',
        '02001': 'no_additional_dynamic_result_sets_returned',
        '03000': 'sql_statement_not_yet_complete',
        '08000': 'connection_exception',
        '08003': 'connection_does_not_exist',
        '08006': 'connection_failure',
        '08001': 'sqlclient_unable_to_establish_sqlconnection',
        '08004': 'sqlserver_rejected_establishment_of_sqlconnection',
        '08007': 'transaction_resolution_unknown',
        '08P01': 'protocol_violation',
        '09000': 'triggered_action_exception',
        '0A000': 'feature_not_supported',
        '0B000': 'invalid_transaction_initiation',
        '0F000': 'locator_exception',
        '0F001': 'invalid_locator_specification',
        '0L000': 'invalid_grantor',
        '0LP01': 'invalid_grant_operation',
        '0P000': 'invalid_role_specification',
        '20000': 'case_not_found',
        '21000': 'cardinality_violation',
        '22000': 'data_exception',
        '2202E': 'array_subscript_error',
        '22021': 'character_not_in_repertoire',
        '22008': 'datetime_field_overflow',
        '22012': 'division_by_zero',
        '22005': 'error_in_assignment',
        '2200B': 'escape_character_conflict',
        '22022': 'indicator_overflow',
        '22015': 'interval_field_overflow',
        '2201E': 'invalid_argument_for_logarithm',
        '22014': 'invalid_argument_for_ntile_function',
        '22016': 'invalid_argument_for_nth_value_function',
        '2201F': 'invalid_argument_for_power_function',
        '2201G': 'invalid_argument_for_width_bucket_function',
        '22018': 'invalid_character_value_for_cast',
        '22007': 'invalid_datetime_format',
        '22019': 'invalid_escape_character',
        '2200D': 'invalid_escape_octet',
        '22025': 'invalid_escape_sequence',
        '22P06': 'nonstandard_use_of_escape_character',
        '22010': 'invalid_indicator_parameter_value',
        '22023': 'invalid_parameter_value',
        '2201B': 'invalid_regular_expression',
        '2201W': 'invalid_row_count_in_limit_clause',
        '2201X': 'invalid_row_count_in_result_offset_clause',
        '22009': 'invalid_time_zone_displacement_value',
        '2200C': 'invalid_use_of_escape_character',
        '2200G': 'most_specific_type_mismatch',
        '22004': 'null_value_not_allowed',
        '22002': 'null_value_no_indicator_parameter',
        '22003': 'numeric_value_out_of_range',
        '22026': 'string_data_length_mismatch',
        '22001': 'string_data_right_truncation',
        '22011': 'substring_error',
        '22027': 'trim_error',
        '22024': 'unterminated_c_string',
        '2200F': 'zero_length_character_string',
        '22P01': 'floating_point_exception',
        '22P02': 'invalid_text_representation',
        '22P03': 'invalid_binary_representation',
        '22P04': 'bad_copy_file_format',
        '22P05': 'untranslatable_character',
        '2200L': 'not_an_xml_document',
        '2200M': 'invalid_xml_document',
        '2200N': 'invalid_xml_content',
        '2200S': 'invalid_xml_comment',
        '2200T': 'invalid_xml_processing_instruction',
        '23000': 'integrity_constraint_violation',
        '23001': 'restrict_violation',
        '23502': 'not_null_violation',
        '23503': 'foreign_key_violation',
        '23505': 'unique_violation',
        '23514': 'check_violation',
        '23P01': 'exclusion_violation',
        '24000': 'invalid_cursor_state',
        '25000': 'invalid_transaction_state',
        '25001': 'active_sql_transaction',
        '25002': 'branch_transaction_already_active',
        '25008': 'held_cursor_requires_same_isolation_level',
        '25003': 'inappropriate_access_mode_for_branch_transaction',
        '25004': 'inappropriate_isolation_level_for_branch_transaction',
        '25005': 'no_active_sql_transaction_for_branch_transaction',
        '25006': 'read_only_sql_transaction',
        '25007': 'schema_and_data_statement_mixing_not_supported',
        '25P01': 'no_active_sql_transaction',
        '25P02': 'in_failed_sql_transaction',
        '26000': 'invalid_sql_statement_name',
        '27000': 'triggered_data_change_violation',
        '28000': 'invalid_authorization_specification',
        '28P01': 'invalid_password',
        '2B000': 'dependent_privilege_descriptors_still_exist',
        '2BP01': 'dependent_objects_still_exist',
        '2D000': 'invalid_transaction_termination',
        '2F000': 'sql_routine_exception',
        '2F005': 'function_executed_no_return_statement',
        '2F002': 'modifying_sql_data_not_permitted',
        '2F003': 'prohibited_sql_statement_attempted',
        '2F004': 'reading_sql_data_not_permitted',
        '34000': 'invalid_cursor_name',
        '38000': 'external_routine_exception',
        '38001': 'containing_sql_not_permitted',
        '38002': 'modifying_sql_data_not_permitted',
        '38003': 'prohibited_sql_statement_attempted',
        '38004': 'reading_sql_data_not_permitted',
        '39000': 'external_routine_invocation_exception',
        '39001': 'invalid_sqlstate_returned',
        '39004': 'null_value_not_allowed',
        '39P01': 'trigger_protocol_violated',
        '39P02': 'srf_protocol_violated',
        '3B000': 'savepoint_exception',
        '3B001': 'invalid_savepoint_specification',
        '3D000': 'invalid_catalog_name',
        '3F000': 'invalid_schema_name',
        '40000': 'transaction_rollback',
        '40002': 'transaction_integrity_constraint_violation',
        '40001': 'serialization_failure',
        '40003': 'statement_completion_unknown',
        '40P01': 'deadlock_detected',
        '42000': 'syntax_error_or_access_rule_violation',
        '42601': 'syntax_error',
        '42501': 'insufficient_privilege',
        '42846': 'cannot_coerce',
        '42803': 'grouping_error',
        '42P20': 'windowing_error',
        '42P19': 'invalid_recursion',
        '42830': 'invalid_foreign_key',
        '42602': 'invalid_name',
        '42622': 'name_too_long',
        '42939': 'reserved_name',
        '42804': 'datatype_mismatch',
        '42P18': 'indeterminate_datatype',
        '42P21': 'collation_mismatch',
        '42P22': 'indeterminate_collation',
        '42809': 'wrong_object_type',
        '42703': 'undefined_column',
        '42883': 'undefined_function',
        '42P01': 'undefined_table',
        '42P02': 'undefined_parameter',
        '42704': 'undefined_object',
        '42701': 'duplicate_column',
        '42P03': 'duplicate_cursor',
        '42P04': 'duplicate_database',
        '42723': 'duplicate_function',
        '42P05': 'duplicate_prepared_statement',
        '42P06': 'duplicate_schema',
        '42P07': 'duplicate_table',
        '42712': 'duplicate_alias',
        '42710': 'duplicate_object',
        '42702': 'ambiguous_column',
        '42725': 'ambiguous_function',
        '42P08': 'ambiguous_parameter',
        '42P09': 'ambiguous_alias',
        '42P10': 'invalid_column_reference',
        '42611': 'invalid_column_definition',
        '42P11': 'invalid_cursor_definition',
        '42P12': 'invalid_database_definition',
        '42P13': 'invalid_function_definition',
        '42P14': 'invalid_prepared_statement_definition',
        '42P15': 'invalid_schema_definition',
        '42P16': 'invalid_table_definition',
        '42P17': 'invalid_object_definition',
        '44000': 'with_check_option_violation',
        '53000': 'insufficient_resources',
        '53100': 'disk_full',
        '53200': 'out_of_memory',
        '53300': 'too_many_connections',
        '54000': 'program_limit_exceeded',
        '54001': 'statement_too_complex',
        '54011': 'too_many_columns',
        '54023': 'too_many_arguments',
        '55000': 'object_not_in_prerequisite_state',
        '55006': 'object_in_use',
        '55P02': 'cant_change_runtime_param',
        '55P03': 'lock_not_available',
        '57000': 'operator_intervention',
        '57014': 'query_canceled',
        '57P01': 'admin_shutdown',
        '57P02': 'crash_shutdown',
        '57P03': 'cannot_connect_now',
        '57P04': 'database_dropped',
        '58030': 'io_error',
        '58P01': 'undefined_file',
        '58P02': 'duplicate_file',
        'F0000': 'config_file_error',
        'F0001': 'lock_file_exists',
        'HV000': 'fdw_error',
        'HV005': 'fdw_column_name_not_found',
        'HV002': 'fdw_dynamic_parameter_value_needed',
        'HV010': 'fdw_function_sequence_error',
        'HV021': 'fdw_inconsistent_descriptor_information',
        'HV024': 'fdw_invalid_attribute_value',
        'HV007': 'fdw_invalid_column_name',
        'HV008': 'fdw_invalid_column_number',
        'HV004': 'fdw_invalid_data_type',
        'HV006': 'fdw_invalid_data_type_descriptors',
        'HV091': 'fdw_invalid_descriptor_field_identifier',
        'HV00B': 'fdw_invalid_handle',
        'HV00C': 'fdw_invalid_option_index',
        'HV00D': 'fdw_invalid_option_name',
        'HV090': 'fdw_invalid_string_length_or_buffer_length',
        'HV00A': 'fdw_invalid_string_format',
        'HV009': 'fdw_invalid_use_of_null_pointer',
        'HV014': 'fdw_too_many_handles',
        'HV001': 'fdw_out_of_memory',
        'HV00P': 'fdw_no_schemas',
        'HV00J': 'fdw_option_name_not_found',
        'HV00K': 'fdw_reply_handle',
        'HV00Q': 'fdw_schema_not_found',
        'HV00R': 'fdw_table_not_found',
        'HV00L': 'fdw_unable_to_create_execution',
        'HV00M': 'fdw_unable_to_create_reply',
        'HV00N': 'fdw_unable_to_establish_connection',
        'P0000': 'plpgsql_error',
        'P0001': 'raise_exception',
        'P0002': 'no_data_found',
        'P0003': 'too_many_rows',
        'XX000': 'internal_error',
        'XX001': 'data_corrupted',
        'XX002': 'index_corrupted'
    }



import unittest
from datetime import datetime
from nanohttp import json, settings
from sqlalchemy import Unicode, Integer, DateTime

from restfulpy.controllers import JsonPatchControllerMixin, ModelRestController
from restfulpy.orm import commit, DeclarativeBase, Field, DBSession
from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication


class ModelValidationCheckingModel(DeclarativeBase):
    __tablename__ = 'model_validation_checking_model'

    id = Field(Integer, primary_key=True, readonly=True)
    title = Field(Unicode(50), unique=True, nullable=False, pattern='[A-Z][a-zA-Z]*')
    modified = Field(DateTime, default=datetime.now, readonly=True)


class Root(JsonPatchControllerMixin, ModelRestController):
    __model__ = ModelValidationCheckingModel

    @ModelValidationCheckingModel.validate
    @json
    @commit
    def post(self):
        m = ModelValidationCheckingModel()
        m.update_from_request()
        DBSession.add(m)
        return m

    @json
    @ModelValidationCheckingModel.expose
    def get(self, title: str=None):
        query = ModelValidationCheckingModel.query
        if title:
            return query.filter(ModelValidationCheckingModel.title == title).one_or_none()
        return query


class ModelValidationDecoratorTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root())
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_model_validate_decorator(self):
        # Required
        self.request('ALL', 'POST', '/', params={}, expected_status=400)

        # Correct pattern
        resp, ___ = self.request('ALL', 'POST', '/', params=dict(title='Test'), doc=False)
        self.assertEqual(resp['title'], 'Test')

        # Invalid pattern
        self.request(
            'ALL', 'POST', '/',
            params=dict(title='startWithSmallCase'),
            expected_status=400,
            doc=False
        )

        # Readonly
        self.request(
            'ALL', 'POST', '/',
            params=dict(
                title='Test',
                modified=datetime.now().isoformat()
            ),
            expected_status=400,
            doc=False
        )


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

from nanohttp import json, settings

from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication
from restfulpy.validation import prevent_form
from restfulpy.controllers import RestController, RootController


class ValidationPreventFormController(RestController):
    @json
    @prevent_form
    def post(self):
        return dict()


class Root(RootController):
    validation = ValidationPreventFormController()


class ValidationPreventFormTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root())

    @classmethod
    def configure_app(cls):
        super().configure_app()
        settings.merge("""
            logging:
              loggers:
                default:
                  level: info
            """)

    def test_validation_prevent_form(self):
        # Good
        self.request('All', 'POST', '/validation', doc=False)

        # Bad
        self.request(
            'All', 'POST', '/validation', doc=False,
            params={'param': 'param'},
            expected_status=400
        )
        self.request(
            'All', 'POST', '/validation', doc=False,
            query_string={'param': 'param'},
            expected_status=400
        )


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

from nanohttp import json, settings
from sqlalchemy import Unicode, Integer

from restfulpy.controllers import JsonPatchControllerMixin, ModelRestController
from restfulpy.orm import commit, DeclarativeBase, Field, DBSession, FilteringMixin, PaginationMixin, OrderingMixin, \
    ModifiedMixin
from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication


class SqlErrorCheckingModel(ModifiedMixin, FilteringMixin, PaginationMixin, OrderingMixin, DeclarativeBase):
    __tablename__ = 'sql_error_checking_model'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50), unique=True, nullable=False)


class Root(JsonPatchControllerMixin, ModelRestController):
    __model__ = SqlErrorCheckingModel

    @json
    @commit
    def post(self):
        m = SqlErrorCheckingModel()
        m.update_from_request()
        DBSession.add(m)
        return m

    @json
    @SqlErrorCheckingModel.expose
    def get(self, title: str=None):
        query = SqlErrorCheckingModel.query
        if title:
            return query.filter(SqlErrorCheckingModel.title == title).one_or_none()
        return query


class SqlExceptionsTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root())
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_sql_errors(self):
        resp, ___ = self.request('ALL', 'POST', '/', params=dict(title='test'), doc=False)
        self.assertEqual(resp['title'], 'test')

        # Posting again to get Sql Error
        resp, headers = self.request('ALL', 'POST', '/', params=dict(title='test'), expected_status=409, doc=False)
        self.assertEqual(resp['message'], 'unique_violation')
        self.assertEqual(headers['X-Reason'], '23505')

        # null constraint
        resp, headers = self.request('ALL', 'POST', '/', expected_status=400, doc=False)
        self.assertEqual(resp['message'], 'not_null_violation')
        self.assertEqual(headers['X-Reason'], '23502')


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import functools
from datetime import datetime, date, time
from decimal import Decimal

from nanohttp import context, HttpNotFound, HttpBadRequest
from sqlalchemy import Column, event
from sqlalchemy.orm import validates, Query, CompositeProperty
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.relationships import RelationshipProperty
from sqlalchemy.ext.hybrid import HYBRID_PROPERTY
from sqlalchemy.ext.associationproxy import ASSOCIATION_PROXY
from sqlalchemy.inspection import inspect
from sqlalchemy.sql.sqltypes import NullType

from restfulpy.utils import format_iso_datetime, format_iso_time, to_camel_case
from restfulpy.orm.mixines import PaginationMixin, FilteringMixin, OrderingMixin
from restfulpy.orm.field import Field
from restfulpy.orm.metadata import MetadataField
from restfulpy.validation import validate_form
from restfulpy.constants import ISO_DATETIME_FORMAT, ISO_DATE_FORMAT, ISO_DATETIME_PATTERN


class BaseModel(object):
    _key_serializer = None

    @classmethod
    def get_column(cls, column):
        if isinstance(column, str):
            mapper = inspect(cls)
            return mapper.columns[column]
        # Commented-out by vahid, because I cannot reach here by tests, I think it's not necessary at all.
        # if isinstance(column, SynonymProperty):
        #     return column.parent.columns[column.name]
        return column

    @classmethod
    def import_value(cls, column, v):
        c = cls.get_column(column)
        if isinstance(c, Column) or isinstance(c, InstrumentedAttribute):
            if c.type.python_type is bool and not isinstance(v, bool):
                return str(v).lower() == 'true'
        return v

    @classmethod
    def prepare_for_export(cls, column, v):
        param_name = column.info.get('json') or to_camel_case(column.key)

        if hasattr(column, 'property') and isinstance(column.property, RelationshipProperty) and column.property.uselist:
            result = [c.to_dict() for c in v]

        elif hasattr(column, 'property') and isinstance(column.property, CompositeProperty):
            result = v.__composite_values__()

        elif v is None:
            result = v

        elif isinstance(v, datetime):
            result = format_iso_datetime(v)

        elif isinstance(v, date):
            result = v.isoformat()

        elif isinstance(v, time):
            result = format_iso_time(v)

        elif hasattr(v, 'to_dict'):
            result = v.to_dict()

        elif isinstance(v, Decimal):
            result = str(v)

        # Commented-out by vahid, I think it's not necessary at all.
        # elif isinstance(v, set):
        #     result = list(v)

        else:
            result = v

        return param_name, result

    @classmethod
    def iter_metadata_fields(cls):
        for c in cls.iter_json_columns(relationships=True, include_readonly_columns=True, include_protected_columns=True):
            yield from MetadataField.from_column(cls.get_column(c), info=c.info)

    @classmethod
    def json_metadata(cls):
        fields = {f.json_name: f.to_json() for f in cls.iter_metadata_fields()}
        mapper = inspect(cls)
        return {
            'name': cls.__name__,
            'primaryKeys': [c.key for c in mapper.primary_key],
            'fields': fields
        }

    def update_from_request(self):
        for column, value in self.extract_data_from_request():
            setattr(
                self,
                column.key[1:] if column.key.startswith('_') else column.key,
                self.import_value(column, value)
            )

    @classmethod
    def iter_columns(cls, relationships=True, synonyms=True, composites=True, use_inspection=True, hybrids=True):
        if use_inspection:
            mapper = inspect(cls)
            for k, c in mapper.all_orm_descriptors.items():

                if k == '__mapper__':
                    continue
                    
                if c.extension_type == ASSOCIATION_PROXY:
                    continue
                    
                if (not hybrids and c.extension_type == HYBRID_PROPERTY) \
                        or (not relationships and k in mapper.relationships) \
                        or (not synonyms and k in mapper.synonyms) \
                        or (not composites and k in mapper.composites):
                    continue

                yield getattr(cls, k)

        else:
            # noinspection PyUnresolvedReferences
            for c in cls.__table__.c:
                yield c

    @classmethod
    def iter_json_columns(cls, include_readonly_columns=True, include_protected_columns=False, **kw):
        for c in cls.iter_columns(**kw):
            if (not include_protected_columns and c.info.get('protected')) or \
                    (not include_readonly_columns and c.info.get('readonly')):
                continue

            yield c

    @classmethod
    def extract_data_from_request(cls):
        for c in cls.iter_json_columns(include_protected_columns=True, include_readonly_columns=False):
            param_name = c.info.get('json', to_camel_case(c.key))

            # Commented-out by Vahid, I think it's not necessary at all.
            # if c.info.get('readonly') and param_name in context.form:
            #     if c.info.get('strict', None):
            #         raise HttpBadRequest('Invalid parameter: %s' % c.info['json'])
            #     else:
            #         continue
            if param_name in context.form:
                value = context.form[param_name]
                # Ensuring the python type, and ignoring silently if python type is not specified
                try:
                    c.type.python_type
                except NotImplementedError:
                    yield c, value
                    continue

                if c.type.python_type == datetime:
                    try:
                        match = ISO_DATETIME_PATTERN.match(value)
                        if not match:
                            raise ValueError()
                        extracted_value = datetime.strptime(match.groups()[0], ISO_DATETIME_FORMAT)
                        yield c, extracted_value
                    except ValueError:
                        raise HttpBadRequest('Invalid datetime format')

                elif c.type.python_type == date:
                    try:
                        yield c, datetime.strptime(value, ISO_DATE_FORMAT)
                    except ValueError:
                        raise HttpBadRequest('Invalid date format')
                else:
                    yield c, value

    def to_dict(self):
        result = {}
        for c in self.iter_json_columns():
            result.setdefault(*self.prepare_for_export(c, getattr(self, c.key)))
        return result

    @classmethod
    def create_sort_criteria(cls, sort_columns):
        criteria = []
        for c in cls.iter_json_columns():
            json_name = c.info.get('json', to_camel_case(c.key))
            if json_name in sort_columns:
                criteria.append((c, sort_columns[json_name] == 'desc'))
        return criteria

    # noinspection PyUnresolvedReferences
    @classmethod
    def filter_paginate_sort_query_by_request(cls, query=None):
        query = query or cls.query

        if issubclass(cls, FilteringMixin):
            query = cls.filter_by_request(query)

        if issubclass(cls, OrderingMixin):
            query = cls.sort_by_request(query)

        if issubclass(cls, PaginationMixin):
            query = cls.paginate_by_request(query=query)

        return query

    @classmethod
    def dump_query(cls, query=None):
        result = []
        for o in cls.filter_paginate_sort_query_by_request(query):
            result.append(o.to_dict())
        return result

    @classmethod
    def expose(cls, func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            if result is None:
                raise HttpNotFound()
            if isinstance(result, Query):
                return cls.dump_query(result)
            return result

        return wrapper

    @classmethod
    def create_validation_rules(cls, **rules):
        patterns = {}
        blacklist = []
        requires = []
        for field in cls.iter_metadata_fields():
            if field.pattern:
                patterns[field.json_name] = field.pattern
            if field.readonly:
                blacklist.append(field.json_name)
            elif not field.optional:
                requires.append(field.json_name)
        result = {}
        if patterns:
            result['pattern'] = patterns
        if blacklist:
            result['blacklist'] = blacklist
        if requires:
            result['requires'] = requires
        result.update(rules)
        return result

    @classmethod
    def validate(cls, *args, **kwargs):
        validation_rules = cls.create_validation_rules(**kwargs)
        decorator = validate_form(**validation_rules)
        return decorator(args[0]) if args and callable(args[0]) else decorator


@event.listens_for(BaseModel, 'class_instrument')
def receive_class_instrument(cls):
    for field in cls.iter_columns(relationships=False, synonyms=False, use_inspection=False):
        if not isinstance(field, Field) or not field.can_validate:
            continue
        method_name = 'validate_%s' % field.name
        if not hasattr(cls, method_name):
            def validator(self, key, value):
                return self.get_column(key).validate(value)

            setattr(cls, method_name, validates(field.name)(validator))
